#include "StdAfx.h"
#include "MessagesReader.h"



MessagesReader::MessagesReader()
:iMAPISession_(NULL)
{
}

MessagesReader::~MessagesReader()
{
}

HRESULT MessagesReader::Initialize()
{

    if(FAILED(CoInitializeEx(NULL, 0)))
    {
        return E_FAIL;
    }

    if(MAPIInitialize(NULL) != S_OK)
    {
        return E_FAIL;
    }

    if(MAPILogonEx(0, NULL, NULL, 0, &iMAPISession_) != S_OK)
    {
        MAPIUninitialize();
        return E_FAIL;
    }

    return S_OK;
}

HRESULT MessagesReader::Uninitialize()
{
    iMAPISession_->Logoff(0, 0, 0);
    iMAPISession_->Release();
    iMAPISession_ = NULL;
    MAPIUninitialize();

    return S_OK;
}

HRESULT MessagesReader::AcquireMessages()
{
    HRESULT hr = 0;

    IMAPITable *pIMapiStoresTable = NULL;
    hr = iMAPISession_->GetMsgStoresTable(0, &pIMapiStoresTable);
    if(FAILED(hr))
    {
        return hr;
    }

    ////////////////////////////////////////////////////////////
    // Step 3 - Query the table for the entry that matches the
    // name of the store we are interested in.

    IMsgStore *iMessageStore = NULL;

    SizedSPropTagArray(2, tblColumns) = {2,{PR_DISPLAY_NAME, PR_ENTRYID}};
    pIMapiStoresTable->SetColumns((LPSPropTagArray)&tblColumns, 0);


    ULONG rowCount = GetRowsCount(iMAPISession_);
    bool exit = false;
    while(!exit) 
    {
        SRowSet *pRowSet1 = NULL;

        hr = pIMapiStoresTable->QueryRows(1, 0, &pRowSet1);

        if(SUCCEEDED(hr))
        {   
            if (pRowSet1->cRows != 1)
            {
                exit = true;
                continue;
            }            

            LPWSTR ptrStr = pRowSet1->aRow[0].lpProps[0].Value.lpszW;

            ENTRYID* pEntry =(ENTRYID*)pRowSet1->aRow[0].lpProps[1].Value.bin.lpb;
            ULONG ulStoreBytes = pRowSet1->aRow[0].lpProps[1].Value.bin.cb;

            iMAPISession_->OpenMsgStore(NULL, ulStoreBytes, pEntry, NULL, NULL, &iMessageStore);

            FreeProws(pRowSet1);
            pRowSet1 = NULL;

            HRESULT hRes1 = ReadMessageStore(iMessageStore);

            iMessageStore->Release();
            iMessageStore = NULL;
        }
        else
        {
            exit = true;
            continue;
        }
    };

    // Clean up the store table
    pIMapiStoresTable->Release();


    return S_OK;

}
HRESULT MessagesReader::ReadRecipientNumbers(IMAPITable *pRecipientTable, LPTSTR* pszTo)
{
    HRESULT hr = S_OK;

    if(pRecipientTable != NULL)
    {
        SRowSet * psrs = NULL;
        SPropValue * pspv = NULL;
        SPropValue * pspvLast = NULL;

        //this is used simply to add a ", " after the 1st recipient (in case of multiple recipients)
        BOOL fFirstRecipient =   TRUE;

        size_t pcchA, pcchB;

        //reset passed variable
        size_t len;
        hr = StringCchLength(*pszTo, MAXBUF, &len);
        //CHR(hr);
        ZeroMemory(*pszTo, len);


        while (1)
        {
            // Copy properties to the ADRLIST
            hr = pRecipientTable->QueryRows(1, 0, &psrs);
           
            // Did we hit the end of the table?
            if (psrs->cRows != 1)
                break;

            // Point just past the last property
            pspvLast = psrs->aRow[0].lpProps + psrs->aRow[0].cValues;

            // Loop through all the properties returned for this row
            for (pspv = psrs->aRow[0].lpProps; pspv < pspvLast; ++pspv)
            {
                switch (pspv->ulPropTag)
                {
                    //At this point you may also be interested on PR_DISPLAY_NAME
                case PR_EMAIL_ADDRESS:
                case PR_EMAIL_ADDRESS_A:
                    if (!fFirstRecipient) {
                        hr = StringCchLength(*pszTo, MAXBUF, &pcchA);
                       
                        hr = StringCchCat(*pszTo, pcchA+2+1, TEXT(", "));
                    }
                    else {
                        fFirstRecipient = FALSE;
                    }

                    hr = StringCchLength(*pszTo, MAXBUF, &pcchA);
                    //CHR(hr);

                    hr = StringCchLength(pspv->Value.lpszW, MAXBUF, &pcchB);
                    //CHR(hr);

                    hr = StringCchCat(*pszTo, pcchA+pcchB+1, pspv->Value.lpszW);
                    //CHR(hr);

                    break;
                case PR_DISPLAY_NAME:
                    {
                        int a = 5;
                    }
                    break;
                default:
                    break;
                }
            }

            //Clen rows before re-using
            FreeProws (psrs);
            psrs = NULL;
        }

        //Success!
        hr = S_OK;

    }
    else
    {
        return E_FAIL;
    }


    return hr;
}
HRESULT MessagesReader::ReadMessageProperties(IMessage *pMsg)
{
    const int propertyCount = 11;

    SizedSPropTagArray(propertyCount, rgMsgTags) = {propertyCount,{
        PR_SENDER_NAME, 
            PR_TITLE,
            PR_LAST_MODIFICATION_TIME, 
            PR_MESSAGE_DELIVERY_TIME, 
            PR_DELIVER_TIME,
            PR_MSG_STATUS,
            PR_MESSAGE_FLAGS,
            PR_SUBJECT,
            PR_SUBJECT_PREFIX,
            PR_HASATTACH,
            PR_SENDER_EMAIL_ADDRESS}};
        LPSPropValue rgMsgprops = NULL;

        ULONG cMsgValues = 0;

        HRESULT hr = pMsg->GetProps((LPSPropTagArray)&rgMsgTags, MAPI_UNICODE, &cMsgValues, &rgMsgprops);    

        if (FAILED(hr) || propertyCount != cMsgValues)
        {
            // we can't get all properties.
            return E_FAIL;
        }

        // Here you can get the values from rgMsgprops
        // For example. FILETIME value = rgMsgprops[2].Value.ft;
        //              LPWSTR strValue = rgMsgprops[0].Value.lpszW;
        //              ULONG value = rgMsgprops[5].Value.ul;

        MAPIFreeBuffer(rgMsgprops);

        return hr;
}
HRESULT MessagesReader::ReadRecipientValue(SRowSet * pRowSet)
{
    if (pRowSet == NULL)
    {
        return E_FAIL;
    }
    // Point just past the last property
    SPropValue * pspv = NULL;
    SPropValue * pspvLast = pRowSet->aRow[0].lpProps + pRowSet->aRow[0].cValues;

    MESSAGE_RECIPIENT_ENTRY_FIXED fixedRecipientEntry;
    ::ZeroMemory(&fixedRecipientEntry, sizeof(fixedRecipientEntry));

    // Loop through all the properties returned for this row
    for (pspv = pRowSet->aRow[0].lpProps; pspv < pspvLast; ++pspv)
    {
        switch (pspv->ulPropTag)
        {
            //At this point you may also be interested on PR_DISPLAY_NAME
        case PR_EMAIL_ADDRESS:
        case PR_EMAIL_ADDRESS_A:
            {
                if(pspv->Value.lpszW != NULL)
                {            
                    _tcsncpy(fixedRecipientEntry.Address, pspv->Value.lpszW, sizeof(fixedRecipientEntry.Address) / sizeof(fixedRecipientEntry.Address[0]));
                }
            }
            break;
        case PR_DISPLAY_NAME:
            {
                if(pspv->Value.lpszW != NULL)
                {            
                    _tcsncpy(fixedRecipientEntry.DisplayName, pspv->Value.lpszW, sizeof(fixedRecipientEntry.DisplayName) / sizeof(fixedRecipientEntry.DisplayName[0]));
                }    
            }
            break;
        case PR_RECIPIENT_TYPE:
            {
                fixedRecipientEntry.Type =  pspv->Value.ul;
            }
        default:
            break;
        }
    }
    return S_OK;

}
HRESULT MessagesReader::ReadRecipients(IMessage *pMsg)
{
    HRESULT hr = S_OK;

    //check
    if(pMsg ==  NULL)
    {
        return E_FAIL;
    }

    IMAPITable * pRecipientTable = NULL;
    ULONG recipientCount = 0;

    hr = pMsg->GetRecipientTable(MAPI_UNICODE, &pRecipientTable);

    if(FAILED(hr)) 
    { 
        return hr;
    }

    hr = pRecipientTable->GetRowCount(0, &recipientCount);
    //if(FAILED(hr)) 
    //{
    //    // create the new table for counting rows
    //    IMAPITable * pRecipientTableForCount = NULL;
    //    hr = pMsg->GetRecipientTable(MAPI_UNICODE, &pRecipientTableForCount);
    //    
    //    recipientCount = GetRowsCount(pRecipientTableForCount);
    //    pRecipientTableForCount->Release();
    //}

    if(FAILED(hr)) 
    { 
        return hr;
    }

    for (unsigned long i = 0; i < recipientCount ; ++i)
    {
        SRowSet * pRowSet = NULL;
        hr = pRecipientTable->QueryRows(1, 0, &pRowSet);
        if (FAILED(hr))
        {
            return hr;
        }

        if(pRowSet->cRows != 1)
        {
            break;
        }

        hr = ReadRecipientValue(pRowSet);
        //Clen rows before re-using
        FreeProws(pRowSet);
        pRowSet = NULL;

        if (FAILED(hr))
        {
            return hr;
        }
    }

    pRecipientTable->Release();

    return hr;
}

HRESULT MessagesReader::ReadAttachments(IMessage *pMsg)
{
    HRESULT hr = S_OK;

    //check
    if(pMsg ==  NULL)
    {
        return E_FAIL;
    }

    IMAPITable *attachmentTable = NULL;
    hr = pMsg->GetAttachmentTable(0, &attachmentTable);

    if(FAILED(hr)) 
    { 
        return hr;
    }

    // Get the properties from the table for the attachments

    SizedSPropTagArray(3, tableAttachmentColumns) = {3,{PR_ATTACH_NUM,PR_ATTACH_SIZE,PR_ATTACH_FILENAME}};
    attachmentTable->SetColumns((LPSPropTagArray)&tableAttachmentColumns, 0);

        ULONG attachmentsCount = 0;
        bool exit = false;
        while(!exit) 
        {
            SRowSet *pAttachmentRowSet = NULL;
            hr = attachmentTable->QueryRows(1, 0, &pAttachmentRowSet);
            if(FAILED(hr)) 
            { 
                return hr;
            }
            if(pAttachmentRowSet->cRows != 1)
            {
                FreeProws(pAttachmentRowSet);
                pAttachmentRowSet = NULL;
                exit = true;
                continue;
            }
            ++attachmentsCount;

            
            long attachmentNumber = 0;
            DWORD attachmentSize = 0;
            TCHAR attachmentName[MAX_PATH] = TEXT("\0");

            attachmentNumber = pAttachmentRowSet->aRow[0].lpProps[0].Value.l;
            attachmentSize = pAttachmentRowSet->aRow[0].lpProps[1].Value.ul;
            _tcscpy(attachmentName, pAttachmentRowSet->aRow[0].lpProps[2].
                Value.lpszW);

            FreeProws(pAttachmentRowSet);

            //Get the IAttach interface for the attachment
            IAttach *pAttach = NULL;

            hr = pMsg->OpenAttach(attachmentNumber, 0, 0, &pAttach);
            if(FAILED(hr)) 
            { 
                return hr;
            }

            // Get the IStream interface to stream the PR_ATTACH_DATA_BIN property out. 
            LPSTREAM AttachmentStream = NULL;
            TCHAR filePath[MAX_PATH] = TEXT("\0");

            wsprintf(filePath, TEXT("\\%s"), attachmentName);

            hr = pAttach->OpenProperty(PR_ATTACH_DATA_BIN, NULL, 0, 0,
                (IUnknown **)&AttachmentStream);

            ////////////////////////////////////////////////////////////
            //Read from the stream, write to the file.
            std::vector<unsigned char> outBuffer;

            ReadFromStream(AttachmentStream, outBuffer);

            // Done. Clean up
            AttachmentStream->Release();
            pAttach->Release();
        }


    attachmentTable->Release();

    return hr;

}
HRESULT MessagesReader::ReadFromStream(LPSTREAM pstmBody, std::vector<unsigned char>& outBuffer)
{
    if(pstmBody == NULL)
    {
        return E_FAIL;
    }
    STATSTG stg;
    HRESULT hr = pstmBody->Stat(&stg,STATFLAG_NONAME);
    if (FAILED(hr))
    {
        return hr;
    }

    unsigned int bodysize = stg.cbSize.LowPart;
    if(bodysize == 0)
    {
        return hr;
    }

    unsigned char* bodybuf = new unsigned char[bodysize];
    ZeroMemory(bodybuf, bodysize);
    ULONG read;
    hr = pstmBody->Read(bodybuf, bodysize, &read);
    if (FAILED(hr) || bodysize != read)
    {
        delete bodybuf; 
        return hr;
    }


    outBuffer.assign(bodybuf, bodybuf + bodysize);
    delete [] bodybuf;

    return hr;
}
HRESULT MessagesReader::ReadMimeBody(IMessage *pMsg, std::vector<unsigned char>& outBuffer)
{
    LPSTREAM pstmBody = NULL;
    outBuffer.clear();
    HRESULT hr = pMsg->OpenProperty (PR_CE_MIME_TEXT, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);
    if (hr == MAPI_E_NOT_FOUND)
    {
        return hr;
    }

    ReadFromStream(pstmBody, outBuffer);
    pstmBody->Release();

    return hr;
}


HRESULT MessagesReader::ReadHTMLBody(IMessage *pMsg, std::vector<unsigned char>& outBuffer)
{
    LPSTREAM pstmBody = NULL;
    outBuffer.clear();
    HRESULT hr = pMsg->OpenProperty (0x1013001E/*PR_BODY_HTML_A*/, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);
    if (hr == MAPI_E_NOT_FOUND)
    {
        hr = pMsg->OpenProperty (0x1013001F/*PR_BODY_HTML_W*/, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);
        if (hr == MAPI_E_NOT_FOUND)
        {
            return hr;
        }
    }

    ReadFromStream(pstmBody, outBuffer);
    pstmBody->Release();

    return hr;
}

HRESULT MessagesReader::ReadRawBody(IMessage *pMsg, std::vector<unsigned char>& outBuffer)
{
    LPSTREAM pstmBody = NULL;
    outBuffer.clear();

    HRESULT hr = pMsg->OpenProperty (PR_BODY, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);

    if (hr == MAPI_E_NOT_FOUND)
    {
        hr = pMsg->OpenProperty (PR_BODY_W, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);
        if (hr == MAPI_E_NOT_FOUND)
        {
            hr = pMsg->OpenProperty (PR_BODY_A, NULL, STGM_READ, 0, (IUnknown **) &pstmBody);
            if (hr == MAPI_E_NOT_FOUND)
            {
                //in this case no PR_BODY, nor PR_BODY_W, nor PR_BODY_A

                return hr;
            }
        }
    }

    ReadFromStream(pstmBody, outBuffer);

    pstmBody->Release();

    return hr;
}

HRESULT MessagesReader::ReadMessages(IMsgStore* msgStore, IMAPITable* pSubFolderTable)
{
    HRESULT hr = 0;

    ULONG messageCount = 0;
    hr = pSubFolderTable->GetRowCount(0, &messageCount);
    if(FAILED(hr)) 
    { 
        return E_FAIL;
    }

    for (unsigned long i = 0; i < messageCount ; ++i)
    {
        SRowSet *pRowSet = NULL;        
        SizedSPropTagArray(3, tblMessages) = {3,{PR_SENDER_NAME, PR_SUBJECT, PR_ENTRYID}};
        pSubFolderTable->SetColumns((LPSPropTagArray)&tblMessages, 0);

        hr = pSubFolderTable->QueryRows(1, 0, &pRowSet);

        if(pRowSet->cRows != 1)
        {
            break;
        }

        IMessage *pMsg = NULL;
        hr = msgStore->OpenEntry(pRowSet->aRow[0].lpProps[2].Value.bin.cb, (LPENTRYID)pRowSet->aRow[0].lpProps[2].Value.bin.lpb, NULL, MAPI_MODIFY, NULL, (LPUNKNOWN*)&pMsg);

        FreeProws(pRowSet);


        if(SUCCEEDED(hr)) 
        {         

            hr = ReadMessageProperties(pMsg);

            hr = ReadRecipients(pMsg);

            if(SUCCEEDED(hr)) 
            {     
                std::vector<unsigned char> rawBody;
                hr = ReadRawBody(pMsg, rawBody);

                std::vector<unsigned char> htmlBody;
                hr = ReadHTMLBody(pMsg, htmlBody);

                std::vector<unsigned char> mimeBody;
                hr = ReadMimeBody(pMsg, mimeBody);

            }

            hr = ReadAttachments(pMsg);
            pMsg->Release();
        }		
    }

    return hr;
}

HRESULT MessagesReader::ReadSubFolder(IMsgStore* msgStore, ULONG subFolderIdentifier)
{
    HRESULT hr = 0;

    LPSPropValue rgprops = NULL;
    LPSPropValue lppPropArray = NULL;
    ULONG cValues = 0;
    IMAPIFolder *pSubFolder = NULL;

    // Step #1. read properties from message store.
    SizedSPropTagArray(2, rgTags) = {2,{subFolderIdentifier,PR_OBJECT_TYPE}};

    hr = msgStore->GetProps((LPSPropTagArray)&rgTags, MAPI_UNICODE, &cValues, &rgprops);
    if(FAILED(hr)) 
    {         
        return E_FAIL;
    }

    // Step #2. Open the corresponding folder.
    hr = msgStore->OpenEntry(rgprops[0].Value.bin.cb, (LPENTRYID)rgprops[0].Value.bin.lpb, NULL, MAPI_MODIFY, NULL, (LPUNKNOWN*)&pSubFolder);
    if(FAILED(hr)) 
    {         
        return E_FAIL;
    }
    MAPIFreeBuffer(rgprops);

    // Step #3. Get the table with messages.
    IMAPITable* pSubFolderTable = NULL;			
    hr = pSubFolder->GetContentsTable(0, &pSubFolderTable);

    pSubFolder->Release();


    HRESULT hRes = ReadMessages(msgStore, pSubFolderTable);

    pSubFolderTable->Release();     

    return hr;
}

HRESULT MessagesReader::ReadMessageStore(IMsgStore* msgStore)
{
    HRESULT hr = 0;


    /*
    PR_IPM_OUTBOX_ENTRYID, for the Outbox folder
    PR_IPM_SENTMAIL_ENTRYID, for the Sent Items folder
    PR_IPM_WASTEBASKET_ENTRYID, for the Deleted Items folder
    PR_CE_IPM_DRAFTS_ENTRYID, for the Drafts folder
    PR_CE_IPM_INBOX_ENTRYID, for the Inbox folder
    */

    HRESULT hresInbox  = ReadSubFolder(msgStore, PR_CE_IPM_INBOX_ENTRYID);    // the Inbox folder
    HRESULT hresOutbox = ReadSubFolder(msgStore, PR_IPM_OUTBOX_ENTRYID);      // the Outbox folder
    HRESULT hresSend   = ReadSubFolder(msgStore, PR_IPM_SENTMAIL_ENTRYID);    // the Sent Items folder
    HRESULT hresDrafts = ReadSubFolder(msgStore, PR_CE_IPM_DRAFTS_ENTRYID);   // the Drafts folder
    HRESULT hresBasket = ReadSubFolder(msgStore, PR_IPM_WASTEBASKET_ENTRYID); // the Deleted Items folder
   
    return hr;
}

int MessagesReader::GetRowsCount(IMAPISession *pIMapi)
{

    IMAPITable *pIMapiStoresTable = NULL;
    HRESULT hr = pIMapi->GetMsgStoresTable(0, &pIMapiStoresTable);
    if(FAILED(hr))
    {
        return 0;
    }

    SizedSPropTagArray(2, tblColumns) = {2,{PR_DISPLAY_NAME, PR_ENTRYID}};
    pIMapiStoresTable->SetColumns((LPSPropTagArray)&tblColumns, 0);

    bool exit = false;

    ULONG rowCount = 0;
    while(!exit) 
    {
        SRowSet *pRowSet1 = NULL;
        hr = pIMapiStoresTable->QueryRows(1, 0, &pRowSet1);

        if(SUCCEEDED(hr))
        {   
            if (pRowSet1->cRows != 1)
            {               
                FreeProws(pRowSet1);
                pRowSet1 = NULL;
                exit = true;
                continue;
            }
            rowCount++;
            FreeProws(pRowSet1);
            pRowSet1 = NULL;
        }
    }

    pIMapiStoresTable->Release();
    return rowCount;

}
